# -*- perl -*-
#
# Copyright (C) 1996 Ken MacLeod
# See the file COPYING for distribution terms.
#

use strict;

use Text::GDO::Context;

package Text::GDO::toAscii;

#
# `initted' is true after the first attempt to run `ToAscii', when
# `ToAscii' calls `init'
#
$Text::GDO::toAscii::initted = 0;

$Text::GDO::toAscii::nbsp = "\01";
$Text::GDO::toAscii::emsp = "$Text::GDO::toAscii::nbsp$Text::GDO::toAscii::nbsp";

sub _init {
    my ($prefix, $sdatadir, $file_name);

    foreach $prefix (@::INC) {
	$sdatadir = "$prefix/sdata";
	if (-d $sdatadir) {
	    opendir (SDATADIR, $sdatadir)
		|| die "$::prog: Can't open directory \`$sdatadir' for reading: $!\n";
	    foreach $file_name (readdir (SDATADIR)) {
		next if ($file_name !~ /\.2ab$/);
		open (ENTITYFILE, "$sdatadir/$file_name")
		    || die "$::prog: Can't open \`$sdatadir/$file_name' for reading: $!\n";
		while (<ENTITYFILE>) {
		    chop;
		    s/(^[^\t]+)\t(.*)/$Text::GDO::toAscii::entities{"$1"} = $2/e;
		}
		close (ENTITYFILE);
	    }
	    closedir (SDATADIR);
	}
    }
    # ISOnum
    $Text::GDO::toAscii::entities{"[nbsp  ]"} = "$Text::GDO::toAscii::nbsp";
    # ISOpub
    $Text::GDO::toAscii::entities{"[emsp  ]"} = "$Text::GDO::toAscii::emsp";
    $Text::GDO::toAscii::entities{"[ensp  ]"} = "$Text::GDO::toAscii::nbsp";
    $Text::GDO::toAscii::entities{"[emsp3 ]"} = "$Text::GDO::toAscii::emsp";
    $Text::GDO::toAscii::entities{"[emsp4 ]"} = "$Text::GDO::toAscii::emsp";
    $Text::GDO::toAscii::entities{"[numsp ]"} = "$Text::GDO::toAscii::nbsp";
    $Text::GDO::toAscii::entities{"[puncsp]"} = "$Text::GDO::toAscii::nbsp";
    $Text::GDO::toAscii::entities{"[thinsp]"} = "";
    $Text::GDO::toAscii::entities{"[hairsp]"} = "";
    # XXX fix in sdata
    $Text::GDO::toAscii::entities{"[bull  ]"} = "*";
    $Text::GDO::toAscii::entities{"[tilde ]"} = "~";

    # XXX LinuxDoc, should be fixed in the linuxdoc-dtd patch
    $Text::GDO::toAscii::entities{"</"} = "</";

    $Text::GDO::toAscii::initted = 1;
}

sub stdout {
    my ($self) = @_;

    print $self->{'data'};

    return 1;
}

sub write {
    my ($self, $basename) = @_;

    if (!open (OUTPUT, ">$basename.ascii")) {
	warn "$::prog: could not open $basename.ascii for writing, $!\n";
	return 0;
    }
    print OUTPUT $self->{'data'};
    close (OUTPUT);

    return 1;
}

sub write_handle {
    my ($self, $handle) = @_;

    print $handle $self->{'data'};

    return 1;
}

sub scalar {
    my ($self) = @_;

    return $self->{'data'};
}

package Text::GDO::DocObject;

sub ToAscii {
    my ($self, %args) = @_;
    my ($data);

    Text::GDO::toAscii::_init() if (!$Text::GDO::toAscii::initted);
    my ($context) = new Text::GDO::Context::Ascii (
				      'entities' => \%Text::GDO::toAscii::entities,
				      'pref_compact' => 0,
				      'pref_block' => 'normal',
				      'pref_undertitle' => 1,
				      'rootnode' => $self->rootnode,
				      'current' => [{}]
				      );

    # transfer caller's preferences into context
    my ($key, $value);
    while (($key, $value) = each %args) {
	$context->{$key} = $value;
    }

    $context->push (
		    'before' => 1, 'after' => 1,
		    'indent' => 0, 'rindent' => 0,
		    'line_width' => 76, 'justify' => 'left',
		    'fill' => 1, 'quoting' => 1
		    );
    $data = $self->ascii_to ($context);
    $context->pop;
    $data->[0] = 1
	if ($data->[0] > 1 && $context->{'prefCompact'});
    my ($ascii_obj) = {
	'data' => $data->[2],
	'before' => $data->[0],
	'after' => $data->[1]
	};
    return bless $ascii_obj, 'Text::GDO::toAscii';
}

sub ascii_context {
}

sub ascii_to {
    my ($self, $context) = @_;
    my ($thing, @inlines, @blocks);

    $context->push;
    $self->ascii_context ($context);
    foreach $thing (@{$self->{'content'}}) {
	if ($thing->is_inline) {
	    push (@inlines, $thing);
	} else {
	    if ($#inlines != -1) {
		my ($block) = $self->ascii_format ($context, @inlines);
		push (@blocks, $block)
		    if defined $block;
		@inlines = ();
	    }
	    push (@blocks, $thing->ascii_to($context));
	}
    }
    if ($#inlines != -1) {
	my ($block) = $self->ascii_format ($context, @inlines);
	push (@blocks, $block)
	    if defined $block;
    }
    @blocks = $self->ascii_add_content ($context, @blocks);
    $context->pop;

    return $self->ascii_fill_content ($context, @blocks);
}

sub ascii_add_content {
    my ($self, $context, @blocks) = @_;

    return @blocks;
}

sub ascii_fill_content {
    my ($self, $context, @blocks) = @_;
    
    my ($before) = $context->before;
    my ($after) = $context->after;
    $before = $blocks[0][0] if ($blocks[0][0] > $before);
    $after = $blocks[-1][1] if ($blocks[-1][1] > $after);
    $blocks[0][0] = 0;

    my ($data) = "";
    my ($space_between) = 0;
    my ($block);
    foreach $block (@blocks) {
	# XXX this is a bug.  ascii_format should not be returning an
	# empty block.  I haven't found where this is coming from
	# yet.  The test document is a Snafu5 document,
	# `html-spec-19950922/html-spec.sgm'
	next if ($block->[2] =~ /^\s*$/s);

	$space_between = $block->[0] if ($block->[0] > $space_between);
	$space_between = 1
	    if ($space_between > 1 && $context->{'pref_compact'});
	$data .= "\n" x $space_between;
	$data .= $block->[2];
	$space_between = $block->[1];
    }

    return [$before, $after, $data];
}

sub ascii_format {
    my ($self, $context, @inlines) = @_;
    my ($inline);
    my ($data) = "";

    foreach $inline (@inlines) {
	$data .= $inline->ascii_string ($context);
    }

    if ($context->fill) {
	$data = $self->fmt ($data, $context->indent,
			    $context->rindent,
			    $context->line_width,
			    $context->justify);
    } else {
	# just remove trailing spaces and newlines
	$data =~ s/\s+$//s;
    }
    $data =~ s/$Text::GDO::toAscii::nbsp/ /g;

    if ($data =~ /^\s*$/s) {
	# This handles the case where a newline or whitespace creeps
	# in between two block-like elements while in an element that
	# accepts "PCDATA"
	return undef;
    } else {
	$data .= "\n"
	    if (substr ($data, -1, 1) ne "\n");
	return [$context->before, $context->after, $data];
    }
}

sub ascii_string {
    my ($self, $context) = @_;

    return $self->as_string ($context);
}

package Text::GDO::Inline_Quote;

sub ascii_string {
    my ($self, $context) = @_;
    my ($pre, $post);

    my ($data) = $self->Text::GDO::Inline::ascii_string ($context);

    if ($context->quoting) {
	$pre = $context->charent('[ldquo ]');
	$post = $context->charent('[rdquo ]');
    }

    return join ("", $pre, $data, $post);
}

package Text::GDO::Inline_Literal;

sub ascii_string {
    my ($self, $context) = @_;
    my ($pre, $post);

    my ($data) = $self->Text::GDO::Inline::ascii_string ($context);

    if ($context->quoting) {
	$pre = $context->charent('[lsquo ]');
	$post = $context->charent('[rsquo ]');
    }

    return join ("", $pre, $data, $post);
}

package Text::GDO::Author;

sub ascii_to {
    my ($self, $context) = @_;
    my (@blocks);
    my ($blocking);

    if ($context->{'pref_block'} eq 'normal') {
	$blocking = 'center';
    } else {
	$blocking = 'left';
    }

    $context->push ('before' => 0, 'after' => 0, 'justify' => $blocking);

    my ($name) = $self->name;
    if ($name) {
	push (@blocks,
		 [0, 0, $self->fmt ($name, 0, 0,
				    $context->line_width,
				    $blocking)]);
    }

    push (@blocks, $self->{'title'}->ascii_to ($context))
	if (defined $self->{'title'});

    push (@blocks, $self->{'orgname'}->ascii_to ($context))
	if (defined $self->{'orgname'});

    my ($orgunit);
    foreach $orgunit (@{$self->{'orgunit'}}) {
	push (@blocks, $orgunit->ascii_to ($context));
    }

    my ($address) = $self->address;
    if ($address) {
	push (@blocks,
		 [0, 0, $self->fmt ($address, 0, 0,
				    $context->line_width,
				    $blocking)]);
    }

    push (@blocks, $self->{'email'}->ascii_to ($context))
	if (defined $self->{'email'});

    $context->pop;

    $blocks[0]->[0] = 2; $blocks[-1]->[1] = 2;

    return @blocks;
}

package Text::GDO::Document;

sub ascii_add_content {
    my ($self, $context, @blocks) = @_;
    my ($block, $blocking);

    unshift (@blocks, $self->{'copyright'}->ascii_to ($context))
	if (defined $self->{'copyright'});

    if ($context->{'pref_block'} eq 'normal') {
	$blocking = 'center';
    } else {
	$blocking = 'left';
    }

    if (defined $self->{'abstract'}) {
	$context->push ();
	$context->set_indent (4);
	$context->set_rindent (4);
	unshift (@blocks, $self->{'abstract'}->ascii_to ($context));
	$context->pop;
	unshift (@blocks,
		 [2, 1, $self->fmt ("ABSTRACT", 0, 0,
				    $context->line_width,
				    $blocking)]);
    }

    $context->push ('justify' => $blocking);

    if (defined $self->{'authors'}) {
	my ($author);
	foreach $author (reverse @{$self->{'authors'}}) {
	    unshift (@blocks, $author->ascii_to ($context));
	}
    }

    unshift (@blocks, $self->{'date'}->ascii_to ($context))
	if (defined $self->{'date'});
	

    if (defined $self->{'sub_title'}) {
	$context->push ();
	if ($context->{'pref_block'} eq 'normal') {
	    $context->set_indent (8);
	    $context->set_rindent (8);
	}
	unshift (@blocks, $self->{'sub_title'}->ascii_to ($context));
	$blocks[0]->[0] = 1; $blocks[0]->[1] = 4;
	$context->pop;
    }

    if (defined $self->{'title'}) {
	$context->push ();
	if ($context->{'pref_block'} eq 'normal') {
	    $context->set_indent (8);
	    $context->set_rindent (8);
	}
	unshift (@blocks, $self->{'title'}->ascii_to ($context));
	$blocks[0]->[0] = 7;
	$blocks[0]->[1] = (defined ($self->{'sub_title'}) ? 2 : 4);
	$context->pop;
    }

    $context->pop;

    @blocks;
}

package Text::GDO::Section;

sub ascii_fill_content {
    my ($self, $context, @blocks) = @_;
    my ($block) = $self->Text::GDO::DocObject::ascii_fill_content ($context, @blocks);

    $block->[0] = (4, 3, 2, 1, 1)[$self->level - 1];
    $block->[1] = 1;

    $block;
}

sub ascii_add_content {
    my ($self, $context, @blocks) = @_;
    my ($title);

    my ($sect_num) = $self->formal_number;
    if ($sect_num ne "") {
	# Numbered section
	$sect_num = $self->formal_number . "  ";
    }
    my ($sect_level) = $self->level;

    if (defined $self->{'title'}) {
	$context->push ('indent' => length ($sect_num));
	$title = $self->{'title'}->ascii_to ($context);
	$context->pop;
    } else {
	$title = [1, 1, ""];
    }

    substr ($title->[2], 0, length ($sect_num)) = $sect_num;
    $title->[1] = 0;

    unshift (@blocks, [0, 2, $self->undertitle ($context->line_width,
					 ("*", "=", "-", "-", " ")[$sect_level - 1], $title->[2])])
	if ($context->{'pref_undertitle'});

    unshift (@blocks, $title);

    @blocks;
}

package Text::GDO::Formal;

sub ascii_add_content {
    my ($self, $context, @blocks) = @_;

    if (defined $self->{'title'}) {
	my ($formal_name) = "    " . $self->formal_name . ".  ";
	$context->push ('indent' => length ($formal_name));
	my ($title) = $self->{'title'}->ascii_to ($context);
	$context->pop;

	substr ($title->[2], 0, length ($formal_name)) = $formal_name;
	$title->[1] = 0;

	unshift (@blocks, $title);
    }

    @blocks;
}

package Text::GDO::BlockQuote;

sub ascii_context {
    my ($self, $context) = @_;

    $context->set_indent ($context->indent + 4);
    $context->set_rindent ($context->rindent + 4);
}

package Text::GDO::Line;

sub ascii_fill_content {
    my ($self, $context, @blocks) = @_;
    my ($block) = $self->Text::GDO::DocObject::ascii_fill_content ($context, @blocks);

    $block->[0] = 0;
    $block->[1] = 0;

    return $block;
}

package Text::GDO::Admonition;

sub ascii_add_content {
    my ($self, $context, @blocks) = @_;

    my ($title);
    if (defined $self->{'title'}) {
	$title = $self->{'title'}->ascii_to ($context);
    } else {
	$title = [1, 1, $self->type . "\n"];
    }

    $title->[1] = 0;

    if ($context->{'pref_undertitle'}) {
	unshift (@blocks, [0, 1, $self->undertitle ($context->line_width,
						    "*", $title->[2])]);
	$blocks[0]->[2] = " " x ($context->indent + 4)
	    . $blocks[0]->[2];
    }

    unshift (@blocks, $title);
    $blocks[0]->[2] = " " x ($context->indent + 4)
	. $blocks[0]->[2];

    return @blocks;
}

package Text::GDO::List;

sub ascii_context {
    my ($self, $context) = @_;

    $context->set_list_type ($self->type);
    $context->set_indent ($context->indent + 4);
}

package Text::GDO::ListItem;

sub ascii_fill_content {
    my ($self, $context, @blocks) = @_;
    my ($block) = $self->Text::GDO::DocObject::ascii_fill_content ($context, @blocks);

    SWITCH: {
	($context->list_type eq 'ordered') && do {
	    my ($item_num) = $self->number;
	    my ($str) = "$item_num)";
	    if (length ($str) < 4) {
		$str .= " ";
	    }
	    while (length ($str) < 4) {
		$str = " " . $str;
	    }
	    substr ($block->[2], $context->{'current'}[-2]{'indent'}, 4) = $str;

	    last SWITCH;
	};

	($context->list_type eq 'itemized') && do {
	    substr ($block->[2], $context->{'current'}[-2]{'indent'}, 4) = " *  ";

	    last SWITCH;
	};
    }

    $block->[0] = 0;

    return $block;
}

package Text::GDO::ListTerm;

sub ascii_to {
    my ($self, $context) = @_;
    my (@blocks);

    $context->set_indent ($context->indent - 4);
    my ($block) = $self->Text::GDO::DocObject::ascii_to ($context);
    $context->set_indent ($context->indent + 4);

    $block->[0] = 0; $block->[1] = 0;

    return $block;
}

package Text::GDO::Screen;

sub ascii_context {
    my ($self, $context) = @_;

    $context->set_fill (0);
    $context->set_quoting (0);
}

package Text::GDO::NoFill;

sub ascii_context {
    my ($self, $context) = @_;

    $context->set_fill (0);
}

package Text::GDO::TablePart;

sub ascii_to {
    return Text::GDO::Table::ascii_to (@_);
}

package Text::GDO::Table;

sub ascii_to {
    my ($self, $context) = @_;

    my ($part, $row, $entry);
    my (@avg_length, @min_width, @col_width);
    my ($ii, $jj);

    my ($num_parts) = $self->num_parts;
    my ($num_table_rows) = $self->num_rows;
    my ($num_columns) = $self->num_columns;
    for ($ii = 0; $ii < $num_columns; $ii ++) {
	$avg_length[$ii] = 0;
	$min_width[$ii] = 0;
    }

    #
    # Calculate column widths
    #  1) find minimum word widths for each column (max of any one
    #     word within a column)
    #  2) find average character-length of each column (avg of all
    #     entries in the column)
    #  3) set column width to max of proportion of average column
    #     length and  minimum width (from 1)
    #     a) if proportion is less than minimum, set to minimum and
    #        subtract difference from available space for other columns
    #
    foreach $part ($self->parts) {
	foreach $row ($part->rows) {
	    my ($col_num) = 0;
	    foreach $entry ($row->entries) {
		my ($word);
		my ($test_data) = $entry->ascii_string($context);

		$entry->{'length'} = length ($test_data);
		$avg_length[$col_num] += $entry->{'length'};
		foreach $word (split (/[\s\n]/s, $test_data)) {
		    if (length ($word) > $min_width[$col_num]) {
			$min_width[$col_num] = length ($word);
		    }
		}
		$col_num ++;
	    }
	}
    }

    # make `avg_length' earn it's name :-)
    my ($total_avg_length) = 0;
    for ($jj = 0; $jj < $num_columns; $jj ++) {
	$total_avg_length += ($avg_length[$jj]
			      = ($avg_length[$jj] / $num_table_rows));
    }

    # set `$col_width' to the proportion of the `avg_length' to
    # `total_avg_length', or `$min_width' if that's greater.
    # If `$min_width' _is_ greater, remove that much space from the
    # other columns ($less_space)
    die "$::prog: assert: \$total_avg_length is 0, no table data"
	if ($total_avg_length == 0);
    my ($less_space) = 0;
    # `3' is our inter-column gap
    my ($line_width) = $context->line_width - $num_columns * 3;
    for ($jj = 0; $jj < $num_columns; $jj ++) {
	$col_width[$jj] = int ($line_width * ($avg_length[$jj]/$total_avg_length) + 0.5);
	if ($col_width[$jj] < $min_width[$jj]) {
	    $less_space += $min_width[$jj] - $col_width[$jj];
	    $col_width[$jj] = $min_width[$jj];
	}
    }

    # now that we now how much space is really available, reproportion
    # the space among the wider columns.  If we run into another min
    # greater than column width, then warn
    my ($already_warned) = 0;
    $line_width -= $less_space;
    for ($jj = 0; $jj < $num_columns; $jj ++) {
	if ($col_width[$jj] != $min_width[$jj]) {
	    $col_width[$jj] = int ($line_width * ($avg_length[$jj]/$total_avg_length) + 0.5);
	    if ($col_width[$jj] < $min_width[$jj]) {
		$line_width -= $min_width[$jj] - $col_width[$jj];
		$col_width[$jj] = $min_width[$jj];
		if (!$already_warned) {
		    $already_warned = 1;
		    warn "table too wide\n";
# XXX		    warn "table too wide at line "
#			. $self->{event}->line . "\n";
		}
	    }
	}
    }

    #
    # format every entry of every row of every part
    #

    # our context is pretty bare, `center' is default for ``head'' part
    $context->push ('before' => 1,
		    'after' => 1,
		    'indent' => 0,
		    'rindent' => 0,
		    'line_width' => 0,
		    'justify' => 'center',
		    'fill' => 1);
    my ($data) = "";
    if ($num_parts == 1) {
	#
	# only a ``body'' part, put divider above and set left justify
	#
	my (@entries);
	for ($ii = 0; $ii < $num_columns; $ii ++) {
	    push (@entries, "-" x $col_width[$ii]);
	}
	$data .= join ("   ", @entries) . "\n";
	# `left' is default for other parts
	$context->set_justify ('left');
    }
    foreach $part ($self->parts) {
	my (@rows);
	foreach $row ($part->rows) {
	    my (@entries);
	    my ($col_num) = 0;
	    my ($ascii);
	    foreach $entry ($row->entries) {
		$context->set_line_width ($col_width[$col_num]);
		$ascii = $entry->ascii_to ($context);
		push (@entries, $ascii->[2]);
		$col_num ++;
	    }

	    push (@rows, merge_entries (\@col_width, @entries));
	}
	$data .= join ("\n", @rows);
	my (@entries);
	for ($ii = 0; $ii < $num_columns; $ii ++) {
	    push (@entries, "-" x $col_width[$ii]);
	}
	$data .= join ("   ", @entries) . "\n";
	# `left' is default for other parts
	$context->set_justify ('left');
    }
    $context->pop;

    return [1, 1, $data];
}

sub merge_entries {
    my ($col_widths, @entries) = @_;
    my (@splits, $ii);
    my ($data) = "";

    for ($ii = 0; $ii <= $#{$col_widths}; $ii ++) {
	my (@line_splits) = split (/\n/, $entries[$ii]);
	$splits[$ii] = \@line_splits;
    }

    my ($done) = 0;
    while (!$done) {
	my (@line);
	$done = 1;
	for ($ii = 0; $ii <= $#{$col_widths}; $ii ++) {
	    my ($col_width) = $col_widths->[$ii];
	    if ($ii == $#{$col_widths}) {
		# Leave off extra space at end of line
		$col_width = length ($splits[$ii]->[0]);
	    }
	    if ($#{$splits[$ii]} > -1) {
		push (@line, sprintf ("%-${col_width}.${col_width}s",
				      shift (@{$splits[$ii]})));
	    } else {
		push (@line, " " x $col_width);
	    }
	    ($#{$splits[$ii]} != -1) && ($done = 0);
	}
	$data .= join ("   ", @line) . "\n";
    }

    return $data;
}

package Text::GDO::XRefEnd;

sub ascii_string {
    my ($self, $context) = @_;
    my ($linkend, $id_attr, $title_obj);
    my ($title);

# XXX use title abbreviations?

    $title_obj = $context->{'rootnode'}{'referables'}{$self->{'link'}}{'object'};
#XXX   $titleObj = $context->reference($self->{'link'});
#XXX sigh.  wish I knew how to do "if (defined &$title_obj->formal_name())"
    if (defined $title_obj && eval { $title = $title_obj->formal_name() }) {
	$title = $title . ", " . $title_obj->{'title'}->ascii_string ($context);
	return $title;
    } else {
#XXX in the case of LinuxDoc's LABEL, the DocObject is known, but we
#don't know where it is
	return "[[Unknown link reference]]";
    }
}

package Text::GDO::URL;

sub ascii_string {
    my ($self, $context) = @_;
    # XXX urls and names should be undef not ""
    my ($url) = "";

    if (defined $self->{'url'}) {
	$url = $self->{'url'}->as_string (Text::GDO::Context::string_context());
    }
    my ($name) = $self->Text::GDO::DocObject::ascii_string($context);

    if (($url eq "") || ($url =~ m/mailto:$name/)) {
	return $name;
    } elsif (($name ne "") && ($name ne $url)) {
	return $name . " ($url)";
    } else {
	return $url;
    }
}

package Text::GDO::Data;

# XXX This should parse the data elements into a tokenized form and
# use a TeX-like boxing algorithm.  Prior to filling additional
# characters can be added, then following filling stylized elements
# can be styled.

sub ascii_to {
    my ($self, $context) = @_;
    my ($data);

    $data .= $self->{'source'}->ascii_string ($context);
    if ($context->fill) {
	$data = $self->fmt ($data, $context->indent,
			    $context->rindent,
			    $context->line_width,
			    $context->justify);
    }
    $data =~ s/$Text::GDO::toAscii::nbsp/ /g;

    $data .= "\n"
	if (substr ($data, -1, 1) ne "\n");

    return [$context->before, $context->after, $data];
}

package Text::GDO::SGML::Event;

sub ascii_string {
    my ($event, $context) = @_;

    return $event->as_string ($context);
}

package SCALAR;

sub ascii_to {
    my ($self, $context) = @_;

    return [1, 1, $self];
}

package Text::GDO::DocObject;

#
# fmt re-fills paragraphs (like fmt(1)) in `ascii' output
#

sub _fmt {
    my ($str, $indent, $line_width, $justify) = @_;

    if ($justify eq 'center') {
	$indent += ($line_width - length ($str)) / 2;
    }
    return " " x $indent . $str . "\n";
}

sub fmt {
    my ($self, $str, $indent, $rindent, $line_width, $justify) = @_;

    $str =~ s/[\s\n]+/ /gs;	# strip multiple spaces/newlines
    $str =~ s/^\s//;		# remove leading space
    $str =~ s/\s$//;		# remove trailing space
    $line_width = $line_width - $indent - $rindent;
    $str =~ s/(.{1,$line_width})(\s|$)/&_fmt($1, $indent, $line_width, $justify)/ge;

    return $str;
}

sub undertitle {
    my ($self, $line_width, $repchar, $title) = @_;
    my ($length) = 0;

    $title =~ s/^(.*)$/if (length ($1) > $length) { $length = length ($1);}/mge;
    if ($length > $line_width) {
	$length = $line_width;
    }

    return ($repchar x $length) . "\n";
}

package Text::GDO::Context::Ascii;
@Text::GDO::Context::Ascii::ISA = qw{Text::GDO::Context};

# inherits `fill'
sub before     { return $_[0]->{'current'}[-1]{'before'}; }
sub after      { return $_[0]->{'current'}[-1]{'after'}; }
sub indent     { return $_[0]->{'current'}[-1]{'indent'}; }
sub rindent    { return $_[0]->{'current'}[-1]{'rindent'}; }
sub line_width { return $_[0]->{'current'}[-1]{'line_width'}; }
sub justify    { return $_[0]->{'current'}[-1]{'justify'}; }
sub list_type  { return $_[0]->{'current'}[-1]{'list_type'}; }
sub quoting    { return $_[0]->{'current'}[-1]{'quoting'} }

sub set_before     { return $_[0]->{'current'}[-1]{'before'} = $_[1]; }
sub set_after      { return $_[0]->{'current'}[-1]{'after'} = $_[1]; }
sub set_indent     { return $_[0]->{'current'}[-1]{'indent'} = $_[1]; }
sub set_rindent    { return $_[0]->{'current'}[-1]{'rindent'} = $_[1]; }
sub set_line_width { return $_[0]->{'current'}[-1]{'line_width'} = $_[1]; }
sub set_justify    { return $_[0]->{'current'}[-1]{'justify'} = $_[1]; }
sub set_list_type  { return $_[0]->{'current'}[-1]{'list_type'} = $_[1]; }
sub set_quoting    { return $_[0]->{'current'}[-1]{'quoting'} = $_[1] }

1;
